; ===================================================================
; Project: OutputLib Library
; Description: Provides an interface for creating a two-pane window useful for displaying build and progress output.
; Author: Jason Boggs <vampire DOT valik AT gmail DOT com>
; ===================================================================
#include-once

#Region Members Exported
#cs
	Batch_Main($sProject, Const ByRef $aScripts) - The main function responsible for running the scripts.
	_OutputWindowCreate() - Creates the output window and both output panes.
	_OutputWaitClosed() - When called in the same process as the window exists, waits for the window to be closed, otherwise returns immediately.
	_OutputProgressWrite($sData) - Writes data to the progress pane.
	_OutputBuildWrite($sData) - Writes data to the build pane.
	_OutputBuildWriteLine($sData) - Writes data to the build pane (by entire line).
	_OutputBuildWriteLineError($sData) - Writes data to the build pane (by entire line if error).
	_OutputWindowSetState($iState) - Changes the state of the output window.
	_OutputWindowHandle() - Returns the handle to the output window.
#ce
#EndRegion Members Exported

#Region Includes
#include "GUIRegisterMsgLib.au3"

#include "UtilityLib.au3"
#include <EditConstants.au3>
#include <GUIConstantsEx.au3>
#include <SendMessage.au3>
#include <WindowsConstants.au3>
#EndRegion Includes

#Region Global Variables
Global Enum $OUTPUT_NONE, _
		$OUTPUT_PROGRESS, _
		$OUTPUT_BUILD
Global Const $tagCOPYDATASTRUCT = "ulong_ptr dwData;" & _
		"dword cbData;" & _
		"ptr lpData"
Global Const $g_sOutputWindowTitle = "AutoIt Build Script Output"
Global Const $g_sOutputWindowText = "This is hidden text"

Global $g_hWndOutputWindow = 0, $g_idEditProgress, $g_idEditBuild, $g_idCheckAutoClose, $g_sOutputBuffer = ""
#EndRegion Global Variables

#Region Public Members

#Region Batch_Main()
; ===================================================================
; Batch_Main($sProject, Const ByRef $aScripts)
;
; The main function responsible for running the scripts.
; Parameters:
;	$sProject - IN - The name of the project.
;	$aScripts - IN - An array of scripts to run.
; Returns:
;	0 on success, non-zero on failure.
; ===================================================================
Func Batch_Main($sProject, Const ByRef $aScripts)
	; Create the output window so all output will go into one window.
	_OutputWindowCreate()
	_OutputProgressWrite("==== Starting " & StringTrimRight(@ScriptName, 4) & " (" & $sProject & ") ====" & @CRLF)

	Local $nReturn
	; Run all the scripts
	Local $nSuccess = 0
	For $i = 0 To UBound($aScripts) - 1
		If $aScripts[$i] Then
			_OutputProgressWrite(@CRLF & "Running " & $aScripts[$i] & " (" & $i + 1 & "/" & UBound($aScripts) & ")..." & @CRLF)
			$nReturn = _RunWaitScript(@ScriptDir & "\" & $aScripts[$i], "", @ScriptDir)
			If $nReturn Then ; An error occured.
				_OutputProgressWrite($aScripts[$i] & " did not complete successfully... ")
				Switch MsgBox(262144 + 2 + 8192 + 48, "Warning", $aScripts[$i] & " did not complete successfully.", 300)
					Case -1, 3 ; Abort, assume abort on timeout.
						_OutputProgressWrite("Aborting." & @CRLF)
						$nReturn = 1
						ExitLoop
					Case 4 ; Retry
						_OutputProgressWrite("Retrying." & @CRLF)
						$i -= 1 ; We have to decrement because the For loop will increment
					Case 5 ; Ignore
						_OutputProgressWrite("Ignoring." & @CRLF)
						$nReturn = 0
				EndSwitch
			Else
				$nSuccess += 1
			EndIf
		EndIf
	Next

	$nReturn = UBound($aScripts) - $nSuccess

	; Wait for close (if applicable).
	_OutputProgressWrite("==== " & StringTrimRight(@ScriptName, 4) & " Finished: " & _
			$nSuccess & "/" & UBound($aScripts) & " Completed ====" & @CRLF & @CRLF)
	_OutputWaitClosed($nReturn)

	Return $nReturn
EndFunc   ;==>Batch_Main
#EndRegion Batch_Main()

#Region _OutputWindowCreate()
; ===================================================================
; _OutputWindowCreate()
;
; Creates the output window and both output panes.
; Parameters:
;	None.
; Returns:
;	HWND to the window.
; ===================================================================
Func _OutputWindowCreate()
	Local $iOld = Opt("WinDetectHiddenText", True)
	Local $bExists = WinExists($g_sOutputWindowTitle, $g_sOutputWindowText)
	Opt("WinDetectHiddenText", $iOld)
	If $bExists Then Return 0

	; Variables used to control different aspects of the GUI.
	Local $w = 600, $h = 480
	Local $iFontSize = Default, $iFontWeight = 600
	Local $iEditColor = 0xFFFFFF, $iEditStyle = BitOR($WS_HSCROLL, $WS_VSCROLL, $ES_READONLY)

	$g_hWndOutputWindow = GUICreate($g_sOutputWindowTitle, $w, $h)
	; We use a hidden label with unique test so we can reliably identify the window.
	Local $idLabelHidden = GUICtrlCreateLabel($g_sOutputWindowText, 0, 0, 1, 1)
	GUICtrlSetState($idLabelHidden, $GUI_HIDE)
	Local $idLabelProgress = GUICtrlCreateLabel("Progress Output:", 4, 4, $w - 88, 20)
	$g_idCheckAutoClose = GUICtrlCreateCheckbox("Auto-close?", $w - 80, 4, 76, 20)
	GUICtrlSetFont($idLabelProgress, $iFontSize, $iFontWeight)
	$g_idEditProgress = GUICtrlCreateEdit("", 4, 28, $w - 8, 150, $iEditStyle)
	GUICtrlSetBkColor($g_idEditProgress, $iEditColor)
	__OutputSetEditLimit($g_idEditProgress) ; Max the size of the edit control.

	Local $idLabelBuild = GUICtrlCreateLabel("Build Output:", 4, 186, $w - 8, 20)
	GUICtrlSetFont($idLabelBuild, $iFontSize, $iFontWeight)

	; We define $y explicitly because it's used both for the y position and the height calculation.
	Local $y = 210
	$g_idEditBuild = GUICtrlCreateEdit("", 4, $y, $w - 8, $h - (4 + $y), $iEditStyle)
	GUICtrlSetBkColor($g_idEditBuild, $iEditColor)
	__OutputSetEditLimit($g_idEditBuild) ; Max the size of the edit control.

	_GUIRegisterMsgEx($WM_COPYDATA, __Output_WM_COPYDATA)
	; GUIRegisterMsg($WM_COPYDATA, __Output_WM_COPYDATA)
	GUISetState(@SW_SHOWNORMAL, $g_hWndOutputWindow)
	Return $g_hWndOutputWindow
EndFunc   ;==>_OutputWindowCreate
#EndRegion _OutputWindowCreate()

#Region _OutputWaitClosed()
; ===================================================================
; _OutputWaitClosed()
;
; When called in the same process as the window exists, waits for the window to be closed,
;	otherwise returns immediately.
; Parameters:
;	None.
; Returns:
;	None.
; ===================================================================
Func _OutputWaitClosed($iReturn = 0)
	; Test the state of the Auto-close checkbox.  If the window isn't in this
	; process then this will never be true.
	If BitAND(GUICtrlRead($g_idCheckAutoClose), $GUI_CHECKED) And ($iReturn = 0) Then
		_GUIRegisterMsgEx($WM_COPYDATA, __Output_WM_COPYDATA, True)
		GUIDelete($g_hWndOutputWindow)
	EndIf

	; While the window exists in this process wait for the close event.
	While WinExists(HWnd($g_hWndOutputWindow))
		If GUIGetMsg() = $GUI_EVENT_CLOSE Then
			_GUIRegisterMsgEx($WM_COPYDATA, __Output_WM_COPYDATA, True)
			GUIDelete($g_hWndOutputWindow)
		EndIf
	WEnd
EndFunc   ;==>_OutputWaitClosed
#EndRegion _OutputWaitClosed()

#Region _OutputProgressWrite()
; ===================================================================
; _OutputProgressWrite($sData)
;
; Writes data to the progress pane.
; Parameters:
;	$sData - IN - The data to write.
; Returns:
;	True if successful, False otherwise.
; ===================================================================
Func _OutputProgressWrite($sData)
	Return __OutputWrite($OUTPUT_PROGRESS, $sData)
EndFunc   ;==>_OutputProgressWrite
#EndRegion _OutputProgressWrite()

#Region _OutputBuildWrite()
; ===================================================================
; _OutputBuildWrite($sData)
;
; Writes data to the build pane.
; Parameters:
;	$sData - IN - The data to write.
; Returns:
;	True if successful, False otherwise.
; ===================================================================
Func _OutputBuildWrite($sData)
	Return __OutputWrite($OUTPUT_BUILD, $sData)
EndFunc   ;==>_OutputBuildWrite
#EndRegion _OutputBuildWrite()

#Region _OutputBuildWriteLine()
; ===================================================================
; _OutputBuildWriteLine($sData)
;
; Writes data to the build pane.
; Parameters:
;	$sData - IN - The data to write.
; Returns:
;	True if successful, False otherwise.
; ===================================================================
Func _OutputBuildWriteLine($sData)
	$g_sOutputBuffer &= $sData
	Local $aLine = StringSplit($g_sOutputBuffer, @LF)
	If @error Then Return SetError(@error, 0, False)

	For $i = 1 To $aLine[0] - 1
		_OutputBuildWrite($aLine[$i] & @LF)
	Next

	$g_sOutputBuffer = $aLine[$aLine[0]]
	Return True
EndFunc   ;==>_OutputBuildWriteLine
#EndRegion _OutputBuildWriteLine()

#Region _OutputBuildWriteLineError()
; ===================================================================
; _OutputBuildWriteLineError($sData)
;
; Writes data to the build pane.
; Parameters:
;	$sData - IN - The data to write.
; Returns:
;	True if successful, False otherwise.
; ===================================================================
Func _OutputBuildWriteLineError($sData)
	$g_sOutputBuffer &= $sData
	Local $aLine = StringSplit($g_sOutputBuffer, @LF)
	If @error Then Return SetError(@error, 0, False)

	Local Const $HTML_LENGTH = StringLen("html")
	For $i = 1 To $aLine[0] - 1
		If StringLeft($aLine[$i], $HTML_LENGTH) <> "html" Then _OutputBuildWrite($aLine[$i] & @LF)
	Next

	$g_sOutputBuffer = $aLine[$aLine[0]]
	Return True
EndFunc   ;==>_OutputBuildWriteLineError
#EndRegion _OutputBuildWriteLineError()

#Region _OutputWindowSetState()
; ===================================================================
; _OutputWindowSetState($iState)
;
; Changes the state of the output window.
; Parameters:
;	$iState - IN - An @SW_ state flag to use.
; Returns:
;	None.
; ===================================================================
Func _OutputWindowSetState($iState)
	Local $iOld = Opt("WinDetectHiddenText", True)
	WinSetState($g_sOutputWindowTitle, $g_sOutputWindowText, $iState)
	If $iState = @SW_RESTORE Then WinActivate($g_sOutputWindowTitle, $g_sOutputWindowText)
	Opt("WinDetectHiddenText", $iOld)
EndFunc   ;==>_OutputWindowSetState
#EndRegion _OutputWindowSetState()

#Region _OutputWindowHandle()
; ===================================================================
; _OutputWindowHandle()
;
; Returns the handle to the output window.
; Parameters:
;	None.
; Returns:
;	The handle to the output window.
; ===================================================================
Func _OutputWindowHandle()
	Return WinGetHandle($g_sOutputWindowTitle, $g_sOutputWindowText)
EndFunc   ;==>_OutputWindowHandle
#EndRegion _OutputWindowHandle()

#EndRegion Public Members

#Region Callback Functions

#Region __Output_WM_COPYDATA()
; ===================================================================
; __Output_WM_COPYDATA($hWnd, $iMsg, $wParam, $lParam)
;
; Message handler for WM_COPYDATA which is called when new output needs written.
; Parameters:
;	$hWnd - IN - UNUSED.
;	$iMsg - IN - UNUSED.
;	$wParam - IN - UNUSED.
;	$lParam - IN - Pointer to the $tagCOPYDATASTRUCT.
; Returns:
;	Non-zero if handled, 0 otherwise.
; ===================================================================
Func __Output_WM_COPYDATA($hWnd, $iMsg, $wParam, $lParam)
	#forceref $hWnd, $iMsg, $wParam
	Local $tCOPYDATASTRUCT = DllStructCreate($tagCOPYDATASTRUCT, $lParam)
	Local $tBuffer = DllStructCreate("char cdata[" & DllStructGetData($tCOPYDATASTRUCT, "cbData") & "]", DllStructGetData($tCOPYDATASTRUCT, "lpData"))
	Local $sData = DllStructGetData($tBuffer, "cdata")
	Local $iID = $OUTPUT_NONE
	Switch DllStructGetData($tCOPYDATASTRUCT, "dwData")
		Case $OUTPUT_PROGRESS
			$iID = $g_idEditProgress
		Case $OUTPUT_BUILD
			$iID = $g_idEditBuild
		Case Else
			Return 0
	EndSwitch
	Local $iLen = GUICtrlSendMsg($iID, $WM_GETTEXTLENGTH, 0, 0)
	GUICtrlSendMsg($iID, $EM_SETSEL, $iLen, $iLen)
	GUICtrlSetData($iID, $sData, True)
	Return 1
EndFunc   ;==>__Output_WM_COPYDATA
#EndRegion __Output_WM_COPYDATA()

#EndRegion Callback Functions

#Region Private Members

#Region __OutputSetEditLimit()
; ===================================================================
; __OutputSetEditLimit($idControl)
;
; Sets an edit control's text limit.
; Parameters:
;	$idControl - IN - The handle to the control.
; Returns:
;	None.
; ===================================================================
Func __OutputSetEditLimit($idControl)
	GUICtrlSendMsg($idControl, $EM_LIMITTEXT, 0, 0)
EndFunc   ;==>__OutputSetEditLimit
#EndRegion __OutputSetEditLimit()

#Region __OutputWrite()
; ===================================================================
; __OutputWrite($iOutput, Const ByRef $sData)
;
; Handles writing to an output pane by taking care of the necessary communication logic.
; Parameters:
;	$iOutput - IN - An $OUTPUT_ flag.
;	$sData - IN - The data to write.
; Returns:
;	True if successful, False otherwise.
; ===================================================================
Func __OutputWrite($iOutput, Const ByRef $sMsg)
	; Find the target window.
	Local $iOld = Opt("WinDetectHiddenText", 1)
	Local $hTarget = WinGetHandle($g_sOutputWindowTitle, $g_sOutputWindowText)
	Opt("WinDetectHiddenText", $iOld)
	If Not $hTarget Then Return False

	; Use the hidden window as the source.
	Local $hSource = WinGetHandle(AutoItWinGetTitle())

	; Create the structures
	Local $tCOPYDATASTRUCT = DllStructCreate($tagCOPYDATASTRUCT)
	Local $iSize = StringLen($sMsg) + 1
	Local $tBuffer = DllStructCreate("char cdata[" & $iSize & "]")

	; Fill the structures
	DllStructSetData($tBuffer, "cdata", $sMsg)
	DllStructSetData($tCOPYDATASTRUCT, "dwData", $iOutput)
	DllStructSetData($tCOPYDATASTRUCT, "cbData", DllStructGetSize($tBuffer)) ; $iSize
	DllStructSetData($tCOPYDATASTRUCT, "lpData", DllStructGetPtr($tBuffer))

	; Send the message.
	_SendMessage($hTarget, _
			$WM_COPYDATA, _
			$hSource, _
			DllStructGetPtr($tCOPYDATASTRUCT))
	Return Not @error
EndFunc   ;==>__OutputWrite
#EndRegion __OutputWrite()

#EndRegion Private Members
